import pathlib
import math
import logging
logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
from typing import Optional, Any

from tinytag import TinyTag
from _ma_playback import ffi, lib
from .ma_result import MA_RESULT_STR, MiniaudioError



class Playback:
    """
        A class for controlling playback of a single audio file in a way that's independent of the audio
        file's format. Works for all file formats supported by https://github.com/mackron/miniaudio.
    """
    
    def __init__(self, path_to_file: Optional[str] = ''):
        self.__ma_attrs = ffi.new("Attrs *")
        
        self.__bind(lib.check_available_playback_devices(self.__ma_attrs))

        if self.__ma_attrs.num_playback_devices < 1:
            raise MiniaudioError('No playback device is available!!')
        
        else:
            lib.init_attrs(self.__ma_attrs)
            
            self.__paused: bool = False
            self.__file_duration: float = 0.0

            if path_to_file:
                self.load_file(path_to_file)   
    
    def load_file(self, path_to_file: str) -> None:
        """
            Loads an audio file using one of the available backends.
            This also kills any active playback

        Args:
            path_to_file: The absolute or relative path to the audio file

        Throws a FileNotFoundError if the audio file is not found
        """

        audio_file = pathlib.Path(path_to_file)
        if not audio_file.exists() or not path_to_file:
            raise FileNotFoundError('Audio file not found: {}'.format(path_to_file))
        
        self.__bind(lib.terminate_audio_stream(self.__ma_attrs))

        self.__bind(lib.load_file(self.__ma_attrs, path_to_file.encode('utf-8')))
        self.__bind(lib.init_audio_stream(self.__ma_attrs))
        self.__bind(lib.set_device_volume(self.__ma_attrs))

        self.__file_duration = TinyTag.get(path_to_file).duration

    def play(self) -> None:
        """
            Plays the loaded file from the beginning. Ignores the fact
            that the current playback might be paused. Has no effect if no 
            audio file has been loaded
        """

        if not self.__ma_attrs.audio_stream_ready:
            logging.error('No audio file has been loaded yet!!')
        
        else:
            if self.active:
                self.stop()

            else:
                if self.__ma_attrs.audio_stream_ended_naturally:
                    # the audio file played to completion meanwhile the audio device
                    # is still running so we've got to stop it.

                    self.__bind(lib.stop_audio_stream(self.__ma_attrs))
                    self.__ma_attrs.audio_stream_ended_naturally = False

            self.__ma_attrs.frame_offset = 0
            self.__ma_attrs.frame_offset_modified = True
            self.__paused = False
            self.__bind(lib.start_audio_stream(self.__ma_attrs))

    def stop(self) -> None:
        """
            Stops audio playback regardless of whether or not it is paused. 
            Has no effect if playback is inactive
        """

        if self.active:
            if not self.__paused:
                # only stop the audio stream if self.pause() didn't
                self.__bind(lib.stop_audio_stream(self.__ma_attrs))

            self.__ma_attrs.frame_offset = 0
            self.__paused = False
            
    def pause(self) -> None:
        """
            Pauses audio playback. Has no effect if playback is inactive or is
            already paused
        """

        if self.active and not self.__paused:
            self.__bind(lib.stop_audio_stream(self.__ma_attrs))
            self.__paused = True

    def resume(self) -> None:
        """ 
            Resumes audio playback. Has no effect if playback is inactive or is 
            not paused
        """

        if self.active and self.__paused:
            self.__paused = False
            self.__bind(lib.start_audio_stream(self.__ma_attrs))
    
    def seek(self, pos: float) -> None:
        """
            Moves playback to a specified position. Has no effect if playback is
            inactive
        
        Args:
            pos: position in seconds for playback to jump to. Its value is clamped
                 to the interval [0, self.duration].
        """

        if self.active:
            pos = min(max(pos, 0), self.__file_duration)
            self.__ma_attrs.frame_offset = math.floor(pos * self.__ma_attrs.decoder.outputSampleRate)
            self.__ma_attrs.frame_offset_modified = True
    
    def set_volume(self, volume: float) -> None:
        """
            Sets the volume for the playback device. Persists
            across audio file loads until reset

        Args:
            volume: A value in the interval [0, 1].
        """

        self.__ma_attrs.playback_volume = min(max(volume, 0), 1)
        if self.active:
            self.__bind(lib.set_device_volume(self.__ma_attrs))
    
    def loop_at_end(self, loops_at_end: bool) -> None:
        """
            Set playback to automatically restart after playing to completion. 
            By default this doesn't happen. Persists across audio file loads until reset

        Args:
            loops_at_end: True if playback should loop, False otherwise
        """

        self.__ma_attrs.loops_at_end = loops_at_end
    
    @property
    def active(self) -> bool:
        """
            True if playback is playing or is paused and False otherwise
        """
    
        if not self.__ma_attrs.audio_stream_ready:
            return False
        
        else:
            return self.__ma_attrs.audio_stream_active or self.__paused
    
    @property
    def playing(self) -> bool:
        """
            True if playback is playing
        """

        if not self.__ma_attrs.audio_stream_ready:
            return False
        
        else:
            return self.__ma_attrs.audio_stream_active

    @property
    def curr_pos(self) -> float:
        """
            The playback's current position/offset in seconds from the
            beginning of audio file if playback is active or the file has 
            been loaded and -1 otherwise
        """

        if self.active:
            return self.__ma_attrs.frame_offset / self.__ma_attrs.decoder.outputSampleRate
        
        elif self.__ma_attrs.audio_stream_ready:
            return 0

        else:        
            return -1
    
    @property
    def paused(self) -> bool:
        return self.__paused
    
    @property
    def duration(self) -> float:
        """
            The length in seconds of the audio file, which is 0 if no file has
            been loaded
        """

        return self.__file_duration
    
    @property
    def volume(self) -> float:
        """
            The playback's current volume, a value in the interval [0, 1]
        """

        if self.active:
            self.__bind(lib.get_device_volume(self.__ma_attrs))
        
        return self.__ma_attrs.playback_volume
    
    @property
    def loops_at_end(self) -> bool:
        return self.__ma_attrs.loops_at_end

    def __bind(self, ma_res: int) -> None:
        """ 
            Internal method for checking and throwing possible miniaudio errors. 
            DO NOT CALL
        """

        if ma_res:
            raise MiniaudioError(MA_RESULT_STR[ma_res])

    def __del__(self):
        self.__bind(lib.terminate_audio_stream(self.__ma_attrs))



import os
import platform
from pathlib import Path

from cffi import FFI
ffibuilder = FFI()

ma_defs_path = Path.cwd() / 'just_playback' / 'ma_defs.txt'
with ma_defs_path.open(mode='r') as f:
    ma_defs = f.read()

miniaudio_src = str(Path('just_playback', 'miniaudio', 'miniaudio.c'))
stb_vorbis_src = str(Path('just_playback', 'miniaudio', 'stb_vorbis.c'))
ma_playback_src = str(Path('just_playback', 'ma_playback.c'))
include_dir = str(Path('just_playback'))

libraries = []
compiler_args = []

if os.name == "posix":
    libraries = ["dl", "m", "pthread"]
    compiler_args = ["-g1", "-O3", "-ffast-math"]
    
if platform.system() != "Darwin":
    compiler_args += ["-march=native"]



ffibuilder.cdef( ma_defs + '\n\n'
                """ 
                    typedef struct {  
                        ma_uint32 num_playback_devices;

                        ma_decoder decoder;
                        ma_device_config deviceConfig;
                        ma_device device;

                        ma_uint64 frame_offset;

                        float playback_volume;
                        bool loops_at_end;

                        bool frame_offset_modified;
                        bool audio_stream_ready;
                        bool audio_stream_active;
                        bool audio_stream_ended_naturally;
                    }
                    Attrs;
                    
                    ma_result check_available_playback_devices(Attrs* attrs);
                    void init_attrs(Attrs* attrs);
                    ma_result load_file(Attrs* attrs, const char* path_to_file);
                    ma_result init_audio_stream(Attrs* attrs);
                    ma_result start_audio_stream(Attrs* attrs);
                    ma_result stop_audio_stream(Attrs* attrs);
                    ma_result terminate_audio_stream(Attrs* attrs);
                    void audio_stream_callback(ma_device* pDevice, void* pOutput, const void* pInput, ma_uint32 frameCount);
                    ma_result set_device_volume(Attrs* attrs);
                    ma_result get_device_volume(Attrs* attrs);
                """)

ffibuilder.set_source("_ma_playback",  
            """ 
                    #include "ma_playback.h"
            """,
            sources=[miniaudio_src, stb_vorbis_src, ma_playback_src], 
            include_dirs=[include_dir],  
            libraries=libraries,
            extra_compile_args=compiler_args,
            define_macros=[
                ("MA_NO_GENERATION", "1")
            ]
            )    

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



__all__ = ('MA_RESULT_STR', 'MiniaudioError')


MA_RESULT_STR = dict([
    (0, 'MA_SUCCESS'),
    (-1, 'MA_ERROR'),
    (-2, 'MA_INVALID_ARGS'),
    (-3, 'MA_INVALID_OPERATION'),
    (-4, 'MA_OUT_OF_MEMORY'),
    (-5, 'MA_OUT_OF_RANGE'),
    (-6, 'MA_ACCESS_DENIED'),
    (-7, 'MA_DOES_NOT_EXIST'),
    (-8, 'MA_ALREADY_EXISTS'),
    (-9, 'MA_TOO_MANY_OPEN_FILES'),
    (-10, 'MA_INVALID_FILE'),
    (-11, 'MA_TOO_BIG'),
    (-12, 'MA_PATH_TOO_LONG'),
    (-13, 'MA_NAME_TOO_LONG'),
    (-14, 'MA_NOT_DIRECTORY'),
    (-15, 'MA_IS_DIRECTORY'),
    (-16, 'MA_DIRECTORY_NOT_EMPTY'),
    (-17, 'MA_END_OF_FILE'),
    (-18, 'MA_NO_SPACE'),
    (-19, 'MA_BUSY'),
    (-20, 'MA_IO_ERROR'),
    (-21, 'MA_INTERRUPT'),
    (-22, 'MA_UNAVAILABLE'),
    (-23, 'MA_ALREADY_IN_USE'),
    (-24, 'MA_BAD_ADDRESS'),
    (-25, 'MA_BAD_SEEK'),
    (-26, 'MA_BAD_PIPE'),
    (-27, 'MA_DEADLOCK'),
    (-28, 'MA_TOO_MANY_LINKS'),
    (-29, 'MA_NOT_IMPLEMENTED'),
    (-30, 'MA_NO_MESSAGE'),
    (-31, 'MA_BAD_MESSAGE'),
    (-32, 'MA_NO_DATA_AVAILABLE'),
    (-33, 'MA_INVALID_DATA'),
    (-34, 'MA_TIMEOUT'),
    (-35, 'MA_NO_NETWORK'),
    (-36, 'MA_NOT_UNIQUE'),
    (-37, 'MA_NOT_SOCKET'),
    (-38, 'MA_NO_ADDRESS'),
    (-39, 'MA_BAD_PROTOCOL'),
    (-40, 'MA_PROTOCOL_UNAVAILABLE'),
    (-41, 'MA_PROTOCOL_NOT_SUPPORTED'),
    (-42, 'MA_PROTOCOL_FAMILY_NOT_SUPPORTED'),
    (-43, 'MA_ADDRESS_FAMILY_NOT_SUPPORTED'),
    (-44, 'MA_SOCKET_NOT_SUPPORTED'),
    (-45, 'MA_CONNECTION_RESET'),
    (-46, 'MA_ALREADY_CONNECTED'),
    (-47, 'MA_NOT_CONNECTED'),
    (-48, 'MA_CONNECTION_REFUSED'),
    (-49, 'MA_NO_HOST'),
    (-50, 'MA_IN_PROGRESS'),
    (-51, 'MA_CANCELLED'),
    (-52, 'MA_MEMORY_ALREADY_MAPPED'),
    (-53, 'MA_AT_END'),
    (-100, 'MA_FORMAT_NOT_SUPPORTED'),
    (-101, 'MA_DEVICE_TYPE_NOT_SUPPORTED'),
    (-102, 'MA_SHARE_MODE_NOT_SUPPORTED'),
    (-103, 'MA_NO_BACKEND'),
    (-104, 'MA_NO_DEVICE'),
    (-105, 'MA_API_NOT_FOUND'),
    (-106, 'MA_INVALID_DEVICE_CONFIG'),
    (-107, 'MA_LOOP'),
    (-200, 'MA_DEVICE_NOT_INITIALIZED'),
    (-201, 'MA_DEVICE_ALREADY_INITIALIZED'),
    (-202, 'MA_DEVICE_NOT_STARTED'),
    (-203, 'MA_DEVICE_NOT_STOPPED'),
    (-300, 'MA_FAILED_TO_INIT_BACKEND'),
    (-301, 'MA_FAILED_TO_OPEN_BACKEND_DEVICE'),
    (-302, 'MA_FAILED_TO_START_BACKEND_DEVICE'),
    (-303, 'MA_FAILED_TO_STOP_BACKEND_DEVICE')
])


class MiniaudioError(Exception):
    def __init__(self, *args: object) -> None:
        if args:
            self.message = args[0]
        
        else:
            self.message = None
    
    def __str__(self) -> str:
        return self.message if self.message else 'UNKNOWN MA_ERROR'


from setuptools import setup, find_packages


with open('README.md', encoding='utf-8') as f:
    long_description = f.read().replace('\r\n', '\n')

setup(
    name = 'just_playback',       
    packages = find_packages(),  
    version = '0.1.7',      
    author = 'Cheo Fusi',                  
    author_email = 'fusibrandon13@gmail.com',
    url = 'https://github.com/cheofusi/just_playback', 
    description = 'A small package for playing audio in python \
                    with file format independent methods for playback control', 
    long_description = long_description,
    long_description_content_type='text/markdown',
    license='MIT',   
    keywords = [    
        'audio',  
        'music',
        'wave',
        'wav'
        'mp3',
        'media',
        'song',
        'play', 
        'player',
        'playback', 
        'audioplayer', 
        'mp3player'
    ],
    cffi_modules=["build_ffi_module.py:ffibuilder"],
    install_requires=[            
        "cffi>=1.0.0",
        "tinytag"
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',         
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Topic :: Multimedia",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Sound/Audio :: Players",
        "Topic :: Multimedia :: Sound/Audio :: Players :: MP3",
        'License :: OSI Approved :: MIT License',  
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)